//
//  MCAuthorizationWrapper.m
//  ScheduleHelper
//
//  Created by Michael Ash on 8/28/05.
//  Copyright 2005 Rogue Amoeba. All rights reserved.
//

#import "MCAuthorizationWrapper.h"

#include <Security/Authorization.h>
#include <Security/AuthorizationTags.h>

@interface NSError (ErrorAdditions)

+ (id)sh_errorWithDomain:(NSString *)domain code:(int)code title: (NSString*)title description: (NSString*)frmt, ...;

@end

@implementation NSError (ErrorAdditions)

+ (id)sh_errorWithDomain:(NSString *)domain code:(int)code title: (NSString*)title description: (NSString*)frmt, ...
{
	NSMutableDictionary* userInfo = [NSMutableDictionary dictionaryWithCapacity: 2];
	va_list argList;
	
	if( title )
		[userInfo setObject: title forKey: NSLocalizedDescriptionKey];
	
	if( [frmt length] )
	{
		va_start(argList, frmt);
		frmt = [[NSString alloc] initWithFormat: frmt arguments: argList];
		va_end(argList);
		[userInfo setObject: frmt forKey: @"NSLocalizedRecoverySuggestion"];
		[frmt release];
	}
	
	return [[[self alloc] initWithDomain: domain code: code userInfo: userInfo] autorelease];
}

@end

static char *NonconstCStr( NSString *str )
{
	NSMutableData *data = [NSMutableData dataWithData: [str dataUsingEncoding: NSUTF8StringEncoding]];
	[data setLength: [data length] + 1]; // add terminating NULL
	return [data mutableBytes];
}

BOOL MCRunAuthorizedCommandSequence( NSArray *commands, NSString *prompt, NSError **error )
{
	AuthorizationRef auth;
	OSStatus err;
	unsigned i, j;
	char **commandNames = NULL;
	AuthorizationItem *items = NULL;
	AuthorizationRights rights;
	AuthorizationRights *authorizedRights = NULL;
	
	BOOL success = NO;
	
	AuthorizationEnvironment environment;
	AuthorizationItem environmentItems[1];
	
	environmentItems[0].name = kAuthorizationEnvironmentPrompt;
	environmentItems[0].value = (prompt ? NonconstCStr( prompt ) : "");
	environmentItems[0].valueLength = strlen( environmentItems[0].value );
	environmentItems[0].flags = 0;
	
	environment.count = 1;
	environment.items = environmentItems;
	
	err = AuthorizationCreate( NULL, &environment,
							   kAuthorizationFlagDefaults, &auth );
	if( err != errAuthorizationSuccess)
	{
		if( error )
			*error = [NSError sh_errorWithDomain: NSOSStatusErrorDomain code: err title: NSLocalizedString( @"Authorization failed", @"Authorization errors" ) description: NSLocalizedString( @"AuthorizationCreate returned error %d", @"Authorization errors" ), err];
		goto cleanup;
	}
	
	commandNames = malloc( [commands count] * sizeof( *commandNames ) );
	items = malloc( [commands count] * sizeof( *items ) );
	for( i = 0; i < [commands count]; i++ )
	{
		commandNames[i] = NonconstCStr( [[commands objectAtIndex: i] objectAtIndex: 0] );
		items[i].name = kAuthorizationRightExecute;
		items[i].valueLength = strlen( commandNames[i] );
		items[i].value = commandNames[i];
		items[i].flags = 0;
	}
	
	rights.count = [commands count];
	rights.items = items;
	err = AuthorizationCopyRights( auth, &rights, &environment,
								   kAuthorizationFlagExtendRights | kAuthorizationFlagInteractionAllowed,
								   &authorizedRights );
	if( err != errAuthorizationSuccess )
	{
		if( error )
			*error = [NSError sh_errorWithDomain: NSOSStatusErrorDomain code: err title: NSLocalizedString( @"Authorization failed", @"Authorization errors" ) description: NSLocalizedString( @"AuthorizationCopyRights returned error %d", @"Authorization errors" ), err];
		goto cleanup;
	}
	
	for( i = 0; i < [commands count]; i++ )
	{
		NSArray *argumentsArray = [commands objectAtIndex: i];
		char **arguments = malloc( [argumentsArray count] * sizeof( *arguments ) ); // -1 for command name, +1 for NULL
		FILE *toolPipe = NULL;
		
		for( j = 1; j < [argumentsArray count]; j++ )
			arguments[j - 1] = NonconstCStr( [argumentsArray objectAtIndex: j] );
		arguments[j - 1] = NULL;
		err = AuthorizationExecuteWithPrivileges( auth, commandNames[i], 0, arguments, &toolPipe );
		free( arguments );
		
		if( err != errAuthorizationSuccess )
		{
			if( error )
				*error = [NSError sh_errorWithDomain: NSOSStatusErrorDomain code: err title: NSLocalizedString( @"Authorization failed", @"Authorization errors" ) description: NSLocalizedString( @"AuthorizationExecuteWithPrivileges returned error %d", @"Authorization errors" ), err];
			goto cleanup;
		}
		
		// wait for the subprocess to terminate before moving on
		char buf[128];
		while( !feof( toolPipe ) && !ferror( toolPipe ) )
			fread( buf, 1, sizeof( buf ), toolPipe );
		
		fclose( toolPipe );
	}
	
	success = YES;
		
cleanup:
	if( auth )
		AuthorizationFree( auth, kAuthorizationFlagDestroyRights );
	if( commandNames )
		free( commandNames );
	if( items )
		free( items );
	if( authorizedRights )
		AuthorizationFreeItemSet( authorizedRights );
	
	return success;
}
