/*
 *  MCMemoryStats.c
 *  MemoryCell
 *
 *  Created by Quentin Carnicelli on Thu Jun 10 2004.
 *  Portions Copyright (c) Rogue Amoeba Software, LLC. All Right Reserved.
 *
 *	Note: Some of the code below comes from Darwin, namely most of the 
 *  _read_mem_regions() function. It was originally from the 'top' 
 *  (http://www.opensource.apple.com/darwinsource/10.3.4/top-9/)
 *	and thus this is covered by the Apple Public Source License:
 *
 *  Portions Copyright (c) 1999-2004 Apple Computer, Inc. All Right Reserved.
 * 
 *  This file contains Original Code and/or Modifications of Original Code
 *  as defined in and that are subject to the Apple Public Source License
 *  Version 2.0 (the 'License'). You may not use this file except in
 *  compliance with the License. Please obtain a copy of the License at
 *  http://www.opensource.apple.com/apsl/ and read it before using this
 *  file.
 * 
 *  The Original Code and all software distributed under the License are
 *  distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 *  EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 *  INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 *  Please see the License for the specific language governing rights and
 *  limitations under the License.
 * 
 */

#import "MCMemoryStats.h"
#include <CoreFoundation/CoreFoundation.h>
#include <mach/mach.h>
#include <mach/shared_memory_server.h>

typedef struct
{
	int			obj_id;
	int			size;
	int			share_mode;
	int			resident_page_count;
	int			ref_count;
	int			proc_ref_count;
	int			rb_share_mode;
	vm_size_t	rb_aliased;
	vm_size_t	rb_vprvt;
	vm_size_t	rb_rshrd;
}_process_memory_object;


static _process_memory_object* _aggregate_mem_object( CFMutableDictionaryRef dict, int obj_id, int share_mode, int res_page_count, int ref_count, int size )
{
	CFNumberRef objKey;
	CFMutableDataRef data;
	_process_memory_object* memObj = NULL;
	
	objKey = CFNumberCreate( NULL, kCFNumberIntType, &obj_id );
	if( !objKey ) return NULL;
	
	data = (CFMutableDataRef)CFDictionaryGetValue( dict, objKey );
	if( data )
	{
		memObj = (_process_memory_object*)CFDataGetMutableBytePtr( data );
		memObj->size += size;
		memObj->proc_ref_count++;
	}
	else
	{
		data = CFDataCreateMutable( NULL, sizeof( _process_memory_object ) );
		if( !data ) return NULL;
		
		memObj = (_process_memory_object*)CFDataGetMutableBytePtr( data );
		
		memset( memObj, 0, sizeof( _process_memory_object ) );
		memObj->obj_id = obj_id;
		memObj->size = size;
		memObj->share_mode = share_mode;
		memObj->resident_page_count = res_page_count;
		memObj->ref_count = ref_count;
		memObj->proc_ref_count = 1;

		CFDictionaryAddValue( dict, objKey, data );
		CFRelease( data );
	}

	CFRelease( objKey );

	return memObj;
}

static kern_return_t _read_mem_regions( task_t task, process_memory_stats* outStats )
{
	kern_return_t			err;
	vm_address_t			address;
	vm_size_t				pagesize;
	vm_size_t				size;
	vm_size_t				aliased;
	mach_msg_type_number_t	count;
	mach_port_t				object_name;
	vm_region_top_info_data_t info;
	boolean_t				hasGlobalShared;
	CFMutableDictionaryRef	allMemObjs;
	_process_memory_object* memObj;

	err = host_page_size( mach_host_self(), &pagesize );
	if( err != KERN_SUCCESS ) return err;

	allMemObjs = CFDictionaryCreateMutable( NULL, 0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks );
	if( !allMemObjs ) return KERN_NO_SPACE;

	hasGlobalShared = 0;
	aliased = 0;
	
	for (address=0; 1; address += size)
	{
		count = VM_REGION_TOP_INFO_COUNT;
		err = vm_region(	task,
							&address,
							&size,
							VM_REGION_TOP_INFO,
							(vm_region_info_t)&info,
							&count,
							&object_name);
		if( err != KERN_SUCCESS )
			break; 

		/* Is this region a private shared? */
		if( address >= GLOBAL_SHARED_TEXT_SEGMENT &&
			address < (GLOBAL_SHARED_DATA_SEGMENT + SHARED_DATA_REGION_SIZE)) 
		{
			
			/* Update hasGlobalShared */
			if (!hasGlobalShared && info.share_mode == SM_EMPTY)
			{
				vm_region_basic_info_data_64_t	b_info;

				count = VM_REGION_BASIC_INFO_COUNT_64;
				vm_region_64( task, &address, &size, VM_REGION_BASIC_INFO, (vm_region_info_t)&b_info, &count, &object_name);
				if( err != KERN_SUCCESS)
					break;

				if (b_info.reserved)
					hasGlobalShared = 1;
			}
			
			if( info.share_mode != SM_PRIVATE )
				continue;
		}
		

		switch (info.share_mode)
		{
			case SM_EMPTY:
			break;

			case SM_PRIVATE:
				outStats->fRprvt += info.private_pages_resident * pagesize;
				outStats->fVprvt += size;
			break;

			case SM_COW:
				if (info.ref_count == 1)
				{
					/* Treat as SM_PRIVATE. */
					outStats->fVprvt += size;
					outStats->fRprvt += info.shared_pages_resident * pagesize;
				}
				else
				{
					memObj = _aggregate_mem_object(
								allMemObjs,
								info.obj_id,
								SM_COW,
								info.shared_pages_resident,
								info.ref_count,
								size );
								
					if( !memObj ) return KERN_NO_SPACE;

					/* Roll back, if necessary. */
					if (memObj->proc_ref_count > 1)
					{
						if (memObj->rb_share_mode != SM_EMPTY)
							memObj->share_mode = memObj->rb_share_mode;

						aliased -= memObj->rb_aliased;
						outStats->fVprvt -= memObj->rb_vprvt;
						outStats->fRshrd -= memObj->rb_rshrd;
					}

					/* Clear rollback fields. */
					memObj->rb_share_mode = SM_EMPTY;
					memObj->rb_aliased = 0;
					memObj->rb_vprvt = 0;
					memObj->rb_rshrd = 0;

					if( memObj->share_mode == SM_SHARED && memObj->ref_count == memObj->proc_ref_count )
					{
						memObj->rb_share_mode = memObj->share_mode;
						memObj->share_mode = SM_PRIVATE_ALIASED;
						memObj->rb_aliased += memObj->resident_page_count * pagesize;
						aliased += memObj->resident_page_count * pagesize;
						memObj->rb_vprvt += memObj->size;
						outStats->fVprvt += memObj->size;
					}

					if (memObj->share_mode != SM_PRIVATE_ALIASED)
					{
						memObj->rb_rshrd += memObj->resident_page_count * pagesize;
						outStats->fRshrd += memObj->resident_page_count * pagesize;
					}

					outStats->fVprvt += info.private_pages_resident * pagesize;
				}
				
				outStats->fRprvt += info.private_pages_resident * pagesize;

			break;

			case SM_SHARED:
				/* Insert a record into the memObj hash table. */
				memObj = _aggregate_mem_object(
							allMemObjs,
							info.obj_id,
							SM_SHARED,
							info.shared_pages_resident,
							info.ref_count,
							size );

				if( !memObj ) return KERN_NO_SPACE;

				/* Roll back, if necessary. */
				if (memObj->proc_ref_count > 1)
				{
					if (memObj->rb_share_mode != SM_EMPTY)
						memObj->share_mode = memObj->rb_share_mode;

					aliased -= memObj->rb_aliased;
					outStats->fVprvt -= memObj->rb_vprvt;
					outStats->fRshrd -= memObj->rb_rshrd;
				}

				/* Clear rollback fields. */
				memObj->rb_share_mode = SM_EMPTY;
				memObj->rb_aliased = 0;
				memObj->rb_vprvt = 0;
				memObj->rb_rshrd = 0;

				if (memObj->share_mode == SM_SHARED && memObj->ref_count == memObj->proc_ref_count)
				{
					/* This is a private aliased object. */
					memObj->rb_share_mode = memObj->share_mode;
					memObj->share_mode = SM_PRIVATE_ALIASED;
					memObj->rb_aliased += memObj->resident_page_count * pagesize;
					aliased += memObj->resident_page_count * pagesize;
					memObj->rb_vprvt += memObj->size;
					outStats->fVprvt += memObj->size;
				}

				if (memObj->share_mode != SM_PRIVATE_ALIASED)
				{
					memObj->rb_rshrd += memObj->resident_page_count * pagesize;
					outStats->fRshrd += memObj->resident_page_count * pagesize;
				}

			break;

			default:
				break;
		}
	}

	outStats->fRprvt += aliased;

	if( hasGlobalShared )
	{
		/* Subtract out the globally shared text and data regions. */
		outStats->fVsize -= (SHARED_TEXT_REGION_SIZE + SHARED_DATA_REGION_SIZE);
	}

	CFRelease( allMemObjs );

	return KERN_SUCCESS;
}

static kern_return_t _read_mem_basic_info( task_t task, process_memory_stats* outStats )
{
	kern_return_t			err;
	mach_msg_type_number_t	count;
	task_basic_info_data_t	taskinfo;

	count = TASK_BASIC_INFO_COUNT;
	err = task_info( task, TASK_BASIC_INFO, (task_info_t)&taskinfo, &count );
	if (err != KERN_SUCCESS) return err;
	
	outStats->fRsize = taskinfo.resident_size;
	outStats->fVsize = taskinfo.virtual_size;
	
	return KERN_SUCCESS;
}

static void _compute_mem_changes( const process_memory_stats* oldStats, process_memory_stats* newStats )
{
	newStats->fRsizeChange = (long)newStats->fRsize - oldStats->fRsize;
	newStats->fVsizeChange = (long)newStats->fVsize - oldStats->fVsize;
	newStats->fRprvtChange = (long)newStats->fRprvt - oldStats->fRprvt;
	newStats->fVprvtChange = (long)newStats->fVprvt - oldStats->fVprvt;
	newStats->fRshrdChange = (long)newStats->fRshrd - oldStats->fRshrd;
}

kern_return_t process_get_mem_stats( pid_t pid, const process_memory_stats* lastStats, process_memory_stats* outStats )
{
	kern_return_t			err;
	task_t					task;
	
	if( pid <= 0 )
		return KERN_INVALID_ARGUMENT;
	if( outStats == NULL )
		return KERN_INVALID_ARGUMENT;
	
	memset( outStats, 0, sizeof( process_memory_stats ) );
	
	err = task_for_pid( mach_task_self(), pid, &task );
	if (err != KERN_SUCCESS) return err;
	
	err = _read_mem_basic_info( task, outStats );
	if (err != KERN_SUCCESS) return err;
	
	err = _read_mem_regions( task, outStats );
	if (err != KERN_SUCCESS) return err;

	if( lastStats != nil )
		_compute_mem_changes( lastStats, outStats );

	{
		struct timeval	cur_time;
		
		gettimeofday(&cur_time, NULL);
		
		outStats->fPid = pid;
		outStats->fTimestamp = cur_time.tv_sec;
		outStats->fStatus = kProcessMemoryStatsOK;
	}

	return KERN_SUCCESS;
}

vm_size_t process_mem_stats_get_value( const process_memory_stats* inStats, process_memory_stat_unit inType )
{
	if( !inStats )
		return 0;

	switch( inType )
	{
		case kProcessPrivateMemorySize:
			return inStats->fRprvt;
		case kProcessRealMemorySize:
			return inStats->fRsize;
		case kProcessSharedMemorySize:
			return inStats->fRshrd;
		case kProcessVirtualPrivateMemorySize:
			return inStats->fVprvt;
		case kProcessVirtualMemorySize:
			return inStats->fVsize;
	}
	
	return 0;
}

vm_size_t process_mem_stats_get_value_change( const process_memory_stats* inStats, process_memory_stat_unit inType )
{
	if( !inStats )
		return 0;

	switch( inType )
	{
		case kProcessPrivateMemorySize:
			return inStats->fRprvtChange;
		case kProcessRealMemorySize:
			return inStats->fRsizeChange;
		case kProcessSharedMemorySize:
			return inStats->fRshrdChange;
		case kProcessVirtualPrivateMemorySize:
			return inStats->fVprvtChange;
		case kProcessVirtualMemorySize:
			return inStats->fVsizeChange;
	}
	
	return 0;
}

kern_return_t process_mem_stats_as_data( const process_memory_stats* inStats, CFDataRef* outData )
{
	if( !inStats || !outData )
		return KERN_INVALID_ARGUMENT;

	*outData = CFDataCreate( nil, (void*)inStats, sizeof(process_memory_stats));

	if( !*outData )
		return KERN_NO_SPACE;

	return KERN_SUCCESS;
}

kern_return_t process_mem_stats_from_data( CFDataRef inData, process_memory_stats* outStats )
{
	void* ptr;
	
	if( !inData || !outStats )
		return KERN_INVALID_ARGUMENT;
	
	if( CFDataGetLength( inData ) != sizeof(process_memory_stats) )
		return KERN_INVALID_ARGUMENT;
	
	ptr = (void*)CFDataGetBytePtr( inData );
	if( !ptr )
		return KERN_FAILURE;
	
	memcpy( outStats, ptr, sizeof(process_memory_stats) );

	return KERN_SUCCESS;
}
